#pragma once

#include "../core/utils/utils.h"
#include "../core/debug/evemon.h"
#include "../core/debug/exception.h"
#include "../core/math/symbolic/var.h"
#include "../core/math/optics/paraxial/abcd.h"
#include "../core/math/optics/paraxial/yu.h"
#include "../core/math/optics/paraxial/seidel.h"
#include "../core/math/optics/system/base.h"

/* refractive surface class */
template<typename type> class refractive_surface : public math::optics::system::base<type>
{
public:
	refractive_surface(const math::symbolic::var<type>& radius_curv, const math::symbolic::var<type>& index_before, const math::symbolic::var<type>& index_after)
	{
		this->m_radius_curv = radius_curv;
		this->m_index_before = index_before;
		this->m_index_after = index_after;
	}

	virtual math::optics::paraxial::abcd<type> get_paraxial_matrix(void) const override
	{
		return math::optics::paraxial::abcd<type>(
			math::symbolic::var<type>((type)1),
			math::symbolic::var<type>(),
			-(this->m_index_after - this->m_index_before) / (this->m_index_after * this->m_radius_curv),
			this->m_index_before / this->m_index_after
			);
	}

	virtual std::tuple<math::optics::paraxial::seidel<type>, math::optics::paraxial::yu<type>, math::optics::paraxial::yu<type>> get_seidel_coefficients_ex(const math::optics::paraxial::yu<type>& yu_marginal_in, const math::optics::paraxial::yu<type>& yu_chief_in) const override
	{
		auto M = get_paraxial_matrix();

		auto yu_marginal_out = M * yu_marginal_in;
		auto yu_chief_out = M * yu_chief_in;

		auto seidel_out = math::optics::paraxial::seidel<type>(yu_marginal_in, yu_marginal_out, yu_chief_in, yu_chief_out, (type)1 / this->m_radius_curv, this->m_index_before, this->m_index_after);

		return std::tuple(seidel_out, yu_marginal_out, yu_chief_out);
	}

private:
	math::symbolic::var<type> m_radius_curv, m_index_before, m_index_after;
};